import psutil, random, sys, threading
from subprocess import PIPE
from anvil_downlink_util.pipes import MessagePipe

# State representing which workers are running here
from anvil_downlink_host import send_with_header, maybe_quit_if_draining_and_done, \
    PopenWithGroupKill, get_demote_fn, TIMEOUT, \
    report_worker_stats, BaseWorker

import anvil_downlink_host.full_python.worker_cache as cache


class Worker(BaseWorker):
    def __init__(self, initial_msg, app_version=None, set_timeout=TIMEOUT, task_info=None):
        BaseWorker.__init__(self, initial_msg, task_info)

        first_req_id = initial_msg["id"]

        app_id = initial_msg.get("app-id", "<unknown>")
        self.proc = PopenWithGroupKill([sys.executable, "-um", "anvil_downlink_worker.full_python_worker"],
                                       bufsize=0, stdin=PIPE, stdout=PIPE, preexec_fn=get_demote_fn(app_id))
        self.proc_info = psutil.Process(self.proc.pid)
        self.from_worker = MessagePipe(self.proc.stdout)
        self.to_worker = MessagePipe(self.proc.stdin)

        self.record_inbound_call_started(initial_msg)

        self.timed_out = False
        self.timeout_msg = ""
        self.timeouts = {}
        self.hard_timeout_timer = None
        self.app_version = app_version
        if set_timeout:
            self.set_timeout(first_req_id, set_timeout)

        self.enable_profiling = {first_req_id: initial_msg.get("enable-profiling", False)}
        self.killing_task = False
        self.global_error = None

        threading.Thread(target=self.read_loop).start()

    def set_timeout(self, timeout_key, timeout_duration=TIMEOUT, timeout_msg=""):
        # print("Set timeout for %s, msg %r" % (timeout_key, timeout_msg))
        if timeout_key in self.timeouts:
            return
        timeout_timer = threading.Timer(timeout_duration, lambda: self.soft_timeout(timeout_key, timeout_msg))
        self.timeouts[timeout_key] = timeout_timer
        timeout_timer.start()

    def clear_timeout(self, timeout_key):
        timeout_timer = self.timeouts.pop(timeout_key, None)
        if timeout_timer is not None:
            timeout_timer.cancel()

    def soft_timeout(self, timeout_key, timeout_msg=""):
        # Something has timed out. If we are cached, drain ourselves nicely before terminating
        print("TIMEOUT for %s %r" % (timeout_key, timeout_msg))
        self.timed_out = True
        self.timeout_msg = timeout_msg

        cache.retire_worker(self)
        if len(self.req_ids) == 1:
            self.hard_timeout()


    def hard_timeout(self):
        print("TIMEOUT TERMINATE FOR %s" % self.req_ids)
        self.timed_out = True
        self.proc.terminate()

    def kill_with_error(self, err):
        self.global_error = err
        self.proc.terminate()

    def terminate(self):
        try:
            self.proc.terminate()
            cache.worker_died(self)
        except:
            pass

    def drain(self):
        # Finish up our execution and time out.
        if self.cache_key is None:
            print("Worker for %s is not cached, ignoring drain request" % self.req_ids)
        elif len(self.req_ids) == 0:
            print("Worker for %s drained, terminating instantly (version %s)" % (self.cache_key, self.app_version))
            try:
                self.proc.terminate()
            except:
                pass
        else:
            print("Worker for %s draining, setting timeout (version %s)" % (self.cache_key, self.app_version))
            self.set_timeout("DRAINED")

    def responded(self, req_id):
        self.clear_timeout(req_id)
        self.enable_profiling.pop(req_id, None)
        self.record_inbound_call_complete(req_id)

        #print("Done @%s -> %s" % (self.cache_key, cached_workers.get(self.cache_key)))

    def kill_background_task(self):
        if self.killing_task:
            return
        self.killing_task = True
        print("SOFT KILL BACKGROUND TASK %s" % self.initial_req_id)

        # Request state. If it returns with in 5 seconds, we will die with state, else we hard-kill
        self.hard_timeout_timer = threading.Timer(5, self._hard_kill_background_task)
        self.send({'type': 'GET_TASK_STATE', 'id': 'pre-kill-task-state'})
        self.hard_timeout_timer.start()

    def _hard_kill_background_task(self):
        print("HARD KILL BACKGROUND TASK %s" % self.initial_req_id)
        try:
            self.req_ids.discard('pre-kill-task-state')
            send_with_header({'type': 'NOTIFY_TASK_KILLED', 'id': self.initial_req_id})
        finally:
            self.hard_timeout()

    # Output gets forwarded straight upstream
    def read_loop(self):
        try:
            while True:
                try:
                    msg, bindata = self.from_worker.receive()
                except EOFError:
                    break
                type = msg.get("type")
                id = msg.get("id") or msg.get("requestId")

                if type == "CALL" or type == "GET_APP":
                    self.record_outbound_call_started(msg)
                else:
                    if id is None:
                        if "output" in msg:
                            # Output from unknown thread? Broadcast it.
                            print("Broadcasting output from unknown thread: %s" % msg)
                            for i in self.req_ids:
                                msg["id"] = i
                                send_with_header(msg)
                        else:
                            print("Discarding invalid message with no ID: %s" % repr(msg))
                        continue
                    if id not in self.req_ids and id not in self.outbound_ids:
                        print("Discarding invalid message with bogus ID: %s" % repr(msg))
                        if type == "CHUNK_HEADER":
                            print("Discarding binary data chunk")
                        continue

                try:
                    if type == "CHUNK_HEADER":
                        send_with_header(msg, bindata)
                        if msg.get("lastChunk"):
                            self.transmitted_media(msg['requestId'], msg['mediaId'])
                    else:

                        if "response" in msg and self.enable_profiling.get(id):
                            self.fill_out_profiling(msg)

                        if "response" in msg and msg['id'] == 'pre-kill-task-state':
                            # Special case handling for a "clean" kill (where we manage to recover the state)

                            objects = msg.get('objects', [])
                            for o in objects:
                                if 'path' in o and o['path'][0] == 'response':
                                    o['path'][0] = 'taskState'
                                if 'DataMedia' in o['type']:
                                    msg['objects'] = []
                                    msg['response'] = None
                                    break

                            self.req_ids.discard('pre-kill-task-state')

                            send_with_header({'type': 'NOTIFY_TASK_KILLED', 'id': self.initial_req_id,
                                              'taskState': msg['response'], 'objects': objects})

                            if self.hard_timeout_timer:
                                self.hard_timeout_timer.cancel()

                            self.proc.terminate()
                        else:
                            send_with_header(msg)

                    if "response" in msg or "error" in msg:
                        #if statsd and (id in self.start_times):
                        #    statsd.timing('Downlink.WorkerLifetime', (time.time()*1000) - self.start_times.get(id, 0)*1000)
                        self.on_media_complete(msg, lambda: self.responded(id))
                        if "error" in msg and msg.get("moduleLoadFailed"):
                            self.kill_with_error(msg["error"])


                except UnicodeError:
                    send_with_header({"id": id, "error": {"type": "UnicodeError", "message": "This function returned a binary string (not text). If you want to return binary data, use a BlobMedia object instead."}})
                    self.responded(id)

        except EOFError:
            print("EOFError while reading worker stdout. This should not have happened.")
            pass

        finally:
            self.clean_up_all_outstanding_records()

            rt = self.proc.poll()
            if rt is None:
                self.proc.terminate()
            for _,t in self.timeouts.items():
                t.cancel()

            cache.worker_died(self)

            error_id = "".join([random.choice('0123456789abcdef') for x in range(10)])
            for i in self.req_ids:
                if self.global_error is not None:
                    err = self.global_error
                elif self.timed_out:
                    err = {
                        'message': self.timeout_msg or "Server code took too long",
                        'type': "anvil.server.TimeoutError"
                    }
                elif rt == -9:
                    err = {
                        'message': "Server code execution process was killed. It may have run out of memory: %s" % (error_id),
                        'type': "anvil.server.ExecutionTerminatedError"
                    }
                    sys.stderr.write(err['message'] + " (IDs %s)\n" % i)
                    sys.stderr.flush()
                else:
                    err = {
                        'message': "Server code exited unexpectedly: %s" % (error_id),
                        'type': "anvil.server.ExecutionTerminatedError"
                    }
                    sys.stderr.write(err['message'] + " (IDs %s)\n" % i)
                    sys.stderr.flush()
                send_with_header({'id': i, 'error': err})
            print ("Worker terminated for IDs %s (return code %s) %s" % (self.req_ids, rt, error_id))
            maybe_quit_if_draining_and_done()

    def send(self, msg, bindata=None):
        id = msg.get("id")
        if msg.get("type") in ["CALL", "GET_TASK_STATE", "LAUNCH_REPL", "REPL_COMMAND", "TERMINATE_REPL"]:
            # It's a new request! Start the timeout
            #print ("Setting timeout and routing for new request ID %s" % id)
            self.record_inbound_call_started(msg)
            if msg.get("enable-profiling"):
                self.enable_profiling[id] = True
            if msg["type"] != "REPL_COMMAND":
                self.set_timeout(id)

        elif msg.get("type") == "REPL_KEEPALIVE":
            self.clear_timeout(msg["repl"])
            self.set_timeout(msg["repl"], timeout_msg="Server repl disconnected")
            send_with_header({"id": id, "response": None})
            return

        try:
            self.to_worker.send(msg, bindata)
        except (BrokenPipeError, EOFError) as e:
            print("Host got {}: {} sending to worker, terminating.".format(type(e).__name__, e))
            self.proc.terminate()
            # TODO: send an error back to the client!

        if "response" in msg or "error" in msg:
            self.on_media_complete(msg, lambda: self.record_outbound_call_complete(id))

    def get_task_state(self, msg):
        self.send(msg)

    def handle_inbound_message(self, msg, bindata=None):
        self.send(msg, bindata)
        if bin is not None and msg.get("last_chunk"):
            self.transmitted_media(msg.get("requestId"), msg.get("mediaId"))

    def on_all_inbound_calls_complete(self):
        cache.worker_idle(self)

    report_stats = report_worker_stats



# Run this file, using a serialised version of the RPC protocol on stdin + stdout
from __future__ import absolute_import
import os, sys, threading, json, importlib
from anvil_downlink_worker import handle_incoming_call, load_app
from anvil_downlink_util.pipes import MessagePipe
from anvil import _serialise, _server, _threaded_server
import anvil.server
import anvil.pdf

PIPE_IN = MessagePipe(os.fdopen(os.dup(0), 'rb'))
PIPE_OUT = MessagePipe(os.fdopen(os.dup(1), 'wb'))
OLD_STDERR = os.fdopen(os.dup(2), 'wb')

_threaded_server.console_output = OLD_STDERR

# We use stdin and stdout to talk Anvil-RPC to our manager process.
# We overwrite sys.stdout and sys.stderr with shims that produce 'output' updates across
# that interface, and to make sure no direct use of FDs 0/1/2 get in its way
# To make sure no clever use Overwrite actual stdin and stdout with a thread that forwards them
new_stdout = os.open(os.devnull, os.O_RDWR)
os.dup2(new_stdout, 0)
os.dup2(new_stdout, 1)
#os.dup2(new_stdout, 2)


def write_pipe(data, bin=None):
    try:
        PIPE_OUT.send(data, bin)
    except ValueError:
        raise _server.SerializationError("You can only pass strings, numbers, arrays, lists, LiveObjects and Media to or from server functions")


class DummyStdout:
    def write(self, s):
        write_pipe({'output': s, 'id': _threaded_server.call_info.call_id})
    def flush(self):
        pass


sys.stdout = DummyStdout()
#sys.stderr = sys.stdout

def dbg(s):
    sys.stderr.write(s+"\n")
    sys.stderr.flush()


def send_reqresp(r, collect_capabilities=None, remote_is_trusted=False):
    _serialise.serialise(r, write_pipe, collect_capabilities=collect_capabilities, remote_is_trusted=remote_is_trusted)

_threaded_server.send_reqresp = send_reqresp


def run():
    while True:
        msg, bindata = PIPE_IN.receive()
        type = msg.get("type", None)
        if type in ["CALL", "LAUNCH_BACKGROUND", "LAUNCH_REPL", "REPL_COMMAND", "TERMINATE_REPL"]:
            handle_incoming_call(msg, write_pipe)

        elif type == "PROVIDE_APP":
            load_app(msg["app"])

        elif type == "GET_TASK_STATE":
            def err(*args):
                raise _server.SerializationError("Cannot use BlobMedia objects in task state.")

            try:
                sjson = _server.fill_out_media({'id': msg['id'], 'response': anvil.server.task_state}, err,
                                               remote_is_trusted=False)
                json.dumps(sjson)
            except (TypeError, _server.SerializationError) as e:
                write_pipe({'id': msg['id'], 'error': {'type': 'anvil.server.SerializationError', 'message': "Illegal value in a anvil.server.task_state. " + e.args[0]}})
            except Exception as e:
                write_pipe({'id': msg['id'], 'error': {'type': 'anvil.server.InternalError', 'message': "Could not get task state: " + e.args[0]}})
            else:
                write_pipe(sjson)

        elif type == "CHUNK_HEADER":
            _serialise.process_blob_header(msg)
            _serialise.process_blob(bindata)
        elif type is None and ("response" in msg or "error" in msg):
            _threaded_server.IncomingResponse(msg)
        else:
            print("Downlink worker socket got unrecognised message: "+repr(msg))
            sys.stdout.flush()
            os._exit(1)


if __name__ == "__main__":
    run()



import anvil
import anvil.js
import anvil.server


origin = anvil.server.get_app_origin() or ""


class GoogleSignInButton(anvil.HtmlTemplate):
    _loaded = False
    _path = (
        origin
        + "/_/static/runtime/img/google-signin-buttons/btn.js?sha=277762afd28c6a94830"
    )

    @classmethod
    def load_component(cls):
        if cls._loaded:
            return
        cls._loaded = True
        document = anvil.js.window.document
        s = document.createElement("script")
        s.src = cls._path
        document.head.append(s)

    def __init__(self, text="Sign in with Google"):
        self.load_component()
        anvil.HtmlTemplate.__init__(self, html="<google-signin-button>" + text)



import anvil.server


#!defMethod()!2: "Superclass of all table exceptions" ["__init__"]
#!defClass(anvil.tables,TableError,__builtins__..Exception)!:
class TableError(anvil.server.AnvilWrappedError):
    pass


#!defMethod()!2: "Raised when attempting to accessing a table row that has been deleted - for example, accessing a row after calling its delete() method, or following a link to a deleted row." ["__init__"]
#!defClass(anvil.tables,RowDeleted,anvil.tables.TableError)!:
class RowDeleted(TableError):
    pass


#!defMethod()!2: "Raised when attempting to access a column that does not exist in this table." ["__init__"]
#!defClass(anvil.tables,NoSuchColumnError,anvil.tables.TableError)!:
class NoSuchColumnError(TableError):
    pass


#!defMethod()!2: "Raised when a transaction conflicts and has been aborted." ["__init__"]
#!defClass(anvil.tables,TransactionConflict,anvil.tables.TableError)!:
class TransactionConflict(TableError):
    pass


#!defMethod()!2: "Raised when an app has exceeded its quota." ["__init__"]
#!defClass(anvil.tables,QuotaExceededError,anvil.tables.TableError)!:
class QuotaExceededError(TableError):
    pass


anvil.server._register_exception_type("anvil.tables.TransactionConflict", TransactionConflict)
anvil.server._register_exception_type("anvil.tables.TableError", TableError)
anvil.server._register_exception_type("anvil.tables.RowDeleted", RowDeleted)
anvil.server._register_exception_type("anvil.tables.NoSuchColumnError", NoSuchColumnError)
anvil.server._register_exception_type("anvil.tables.QuotaExceededError", QuotaExceededError)




import anvil, anvil.server
from . import auth

_config = None

def get_config():
	global _config
	if _config is None:
		_config = anvil.server.call("anvil.private.google.get_config")
	return _config



# A stub of the client-side anvil.designer module

def _err_func(*args, **kws):
    raise Exception("You cannot use anvil.designer in server modules")


in_designer = False

get_design_name = _err_func
update_component_properties = _err_func
update_component_sections = _err_func
start_editing_subform = _err_func
start_editing_form = _err_func
register_interaction = _err_func
notify_interactions_changed = _err_func
get_design_component = lambda x: x
request_form_property_change = _err_func
start_inline_editing = _err_func



from anvil_downlink_host import workers_by_id, send_with_header
from . import worker_cache


def launch(data):
    type = data.get("type")
    id = data.get("id")


    is_background_task = type == "LAUNCH_BACKGROUND"
    is_repl_launch = type == "LAUNCH_REPL"

    print ("%s '%s' for app '%s' (ID %s)" % ("Launching REPL" if is_repl_launch else "Launching BG task" if is_background_task else "Calling function",
                                             data.get("command", "<no func>"), data.get("app-id", "<unknown>"), id))

    if data.get('command') == "anvil.private.pdf.get_component" and not is_background_task:
        wid = data['args'][0][0]
        worker = workers_by_id.get(wid)
        if worker is None:
            send_with_header({'id': id, 'error': {'message': "No component worker found for print call '%s'" % wid}})
            return
        else:
            worker.send(data)

    else:
        worker_cache.handle(data)



import os, threading

import anvil_downlink_host
from anvil_downlink_host import TIMEOUT as CALL_TIMEOUT, BACKGROUND_TIMEOUT

CAN_PERSIST = (os.environ.get("DOWNLINK_CAN_PERSIST", "false").lower() in {"true", "1"})
# How long does the latest version of server code stay around? (default: 48 hours)
PRIMARY_PERSIST_TIME = int(os.environ.get("DOWNLINK_PRIMARY_PERSIST_TIMEOUT", 48*60*60))
# How long do other versions of server code (not the latest) stay around? (default: 2 minutes)
SECONDARY_PERSIST_TIME = int(os.environ.get("DOWNLINK_PRIMARY_PERSIST_TIMEOUT", 120))

# cache-key -> app-version -> WorkerEntry
entries_by_cache_key = {}
# Worker -> WorkerEntry
entries_by_worker = {}
CACHE_LOCK = threading.Lock()

# Worker -> kill timer (to cancel if it dies naturally)
retiring_workers = {}


# Half-baked enum for 2.7 compat
class WorkerState:
    NEW = "NEW"
    PRIMARY = "PRIMARY"
    SECONDARY = "SECONDARY"
    DRAINING = "DRAINING"
    RETIRING = "RETIRING"
    DEAD = "DEAD"


class WorkerEntry:
    def __init__(self, worker, cache_key, app_version):
        self.worker = worker
        self.cache_key = cache_key
        self.app_version = app_version
        self.state = WorkerState.NEW
        self.idle = True
        self._n_calls = 0
        self._timeout = None

    def repr(self):
        return "WorkerEntry<%s,%s>" % (self.state, self.cache_key)

    def demote(self):
        # A call has arrived for our cache key but not our version; we are now secondary at best
        if self.state == WorkerState.PRIMARY:
            print("%s demoted" % self)
            self.state = WorkerState.SECONDARY
            self._set_timeout()

    def promote(self):
        # A call has arrived for our cache key and version; we are now primary
        if self.state != WorkerState.PRIMARY:
            print("%s promoted" % self)
        self.state = WorkerState.PRIMARY
        self.idle = False
        self._n_calls += 1
        self._set_timeout()

    def _set_timeout(self):
        self._cancel_timeout()
        if self.state == WorkerState.PRIMARY:
            if PRIMARY_PERSIST_TIME == 0: # infinite is allowed
                return
            n_secs = PRIMARY_PERSIST_TIME
        elif self.state == WorkerState.SECONDARY:
            n_secs = SECONDARY_PERSIST_TIME
        else:
            n_secs = 1 # everyone else dies, like, fast.
        n_calls_at_timeout_set = self._n_calls
        self._timeout = threading.Timer(n_secs, lambda: self._start_draining(n_calls_at_timeout_set))
        self._timeout.start()

    def _cancel_timeout(self):
        if self._timeout is not None:
            self._timeout.cancel()
            self._timeout = None

    def set_dead(self):
        print("%s dead" % self)
        self.state = WorkerState.DEAD
        self._cancel_timeout()

    def set_retiring(self):
        print("%s retiring" % self)
        self.state = WorkerState.RETIRING
        self._cancel_timeout()

    def set_idle(self):
        self.idle = True

    # These are the only methods that may be called *without* CACHE_LOCK held
    def handle_call(self, msg):
        self.worker.send(msg)

    def _start_draining(self, n_calls_at_timeout_set):
        terminate_immediately = False
        with CACHE_LOCK:
            if self._n_calls != n_calls_at_timeout_set:
                # There have been new calls since this timeout, we're irrelevant
                print("%s ignoring drain timeout because we've received calls" % self)
                return
            print("%s draining due to inactivity (idle? %s)" % (self, self.idle))
            self.state = WorkerState.DRAINING
            self._cancel_timeout()
            if self.idle:
                terminate_immediately = True
            else:
                self._timeout = threading.Timer(CALL_TIMEOUT, lambda: self._hard_kill)
                self._timeout.start()
        if terminate_immediately:
            self._hard_kill()

    def _hard_kill(self):
        print("%s hard kill" % self)
        with CACHE_LOCK:
            if self.state != WorkerState.DRAINING:
                # Reprieve!
                print("%s aborted hard kill because no longer draining (now %s)" % (self, self.state))
                return
            _remove_entry_from_cache(self)
            self._cancel_timeout()
        self.worker.terminate()


def _remove_entry_from_cache(entry):
    entries_by_worker.pop(entry.worker, None)
    entries_by_version = entries_by_cache_key.get(entry.cache_key)
    if entries_by_version is not None:
        entries_by_version.pop(entry.app_version, None)
        if not entries_by_version:
            del entries_by_cache_key[entry.cache_key]


def worker_died(worker):
    # This worker is dead; remove it from caches forthwith
    with CACHE_LOCK:
        entry = entries_by_worker.get(worker)
        if entry is not None:
            entry.set_dead()
            _remove_entry_from_cache(entry)
        else:
            retire_timeout = retiring_workers.pop(worker, None)
            print("Uncached worker %s died; retired? %s" % (worker, bool(retire_timeout)))
            if retire_timeout is not None:
                retire_timeout.cancel()


def worker_idle(worker):
    terminate = False
    with CACHE_LOCK:
        entry = entries_by_worker.get(worker)
        if entry is None:
            terminate = True
        elif entry.state == WorkerState.DRAINING:
            _remove_entry_from_cache(entry)
            entry.cancel_timeouts()
            terminate = True
        else:
            entry.set_idle()

    if terminate:
        worker.terminate()


def retire_worker(worker, grace_period=CALL_TIMEOUT):
    """Do not route any more requests to this worker, and kill it if it does not terminate within a grace period"""
    def do_kill():
        retiring_workers.pop(entry.worker, None)
        entry.worker.terminate()

    with CACHE_LOCK:
        entry = entries_by_worker.get(worker)
        if entry:
            print("Retiring %s" % entry)
            entry.set_retiring()
            _remove_entry_from_cache(entry)

            t = retiring_workers[entry.worker] = threading.Timer(grace_period, do_kill)
            t.start()
    if not entry:
        print("Fast kill on timeout")
        worker.terminate()


anvil_downlink_host.retire_cached_worker = retire_worker


def handle(msg):
    # We import Worker in local scope because both 'worker' and 'worker_cache' refer to each other, and some of the
    # obvious ways of doing that don't work in Python 2. Refactor with care.
    from .worker import Worker

    persist_key = msg.get("persist-key")
    app_version = msg.get("app-version")
    app_id = msg.get("app-id")
    bg_task_timeout = msg.get("bg-task-timeout")

    is_background_task = msg["type"] == "LAUNCH_BACKGROUND"
    is_repl_launch = msg["type"] == "LAUNCH_REPL"
    can_timeout = None if is_repl_launch else BACKGROUND_TIMEOUT if is_background_task else CALL_TIMEOUT
    if is_background_task and bg_task_timeout is not None:
        can_timeout = min(bg_task_timeout, can_timeout) if can_timeout else bg_task_timeout
    # print("Timeout?", is_repl_launch, is_background_task, BACKGROUND_TIMEOUT, CALL_TIMEOUT)

    if CAN_PERSIST and not is_background_task and not is_repl_launch and persist_key is not None and app_id and app_version is not None:
        # It's cacheable; let's go.
        cache_key = (app_id, persist_key)

        with CACHE_LOCK:
            entries_by_version = entries_by_cache_key.setdefault(cache_key, {})
            entry = entries_by_version.get(app_version)
            if not entry:
                worker = Worker(msg, app_version=app_version, set_timeout=False, task_info={
                    "app_id": app_id,
                    "type": "persistent_worker",
                    "task": msg.get("command"),
                    "persist": {"key": persist_key, "version": app_version},
                })
                print("Launched new worker: %s" % worker)
                entry = WorkerEntry(worker, cache_key, app_version)
                entries_by_version[app_version] = entry
                entries_by_worker[worker] = entry
            else:
                print("Found worker for %s / %s: %s -> %s" % (cache_key, app_version, entry, entry.worker))
            for e in entries_by_version.values():
                if e is not entry:
                    e.demote()
            entry.promote()

        entry.handle_call(msg)

        #print("Attempt persistence: %s" % cache_key)
        #print("Version %s:\n%s\nvs\n%s" % (("MATCH" if version==supplied_version else "MISMATCH"), version, supplied_version))
    else:
        # Straight launch, no cache
        w = Worker(msg, app_version=app_version, set_timeout=can_timeout, task_info={
            "app_id": app_id,
            "type": "repl" if is_repl_launch else "background_task" if is_background_task else "server_call",
            "task": msg.get("command"),
            "persist": None,
        })
        print("Single-use worker %s" % w)
        w.send(msg)


    # if statsd:
    #     if is_background_task:
    #         statsd.incr('Downlink.LaunchBackgroundTask')
    #     else:
    #         statsd.incr('Downlink.Call')

